/************************************************
Robo CD Changer

Petey the Programmer
Created: 24-May-2009

Version 31.3	Modified: 27-May-2009	(first version = 31.0)	
Version 1.4		Modified 06-June-2009
Version 1.5		Modified 08-June-2009
Version 1.6		Modified 12-June-2009
Version 1.7		Modified 13-June-2009
Version 2.0		Modified 15-June-2009
Version 2.1		Modified 22-June-2009
Version 2.2		Modified 06-May-2010
Version 2.4		Modified 16-May-2010
Version 2.5		Modified 26-May-2010

-- main.c --

Project: Robo CD Changer MPU for ATMega8 chip

	Features: 
		Interupt driven Hardware UART for receiving commands via RS-232 port.
		2 channels of hardware PWM for Motor Direction & Speed Control.
		8 channels of inputs as switches & sensors for position & limit sensing.
		2 12v relays driven by darlington array to operate high current solinoids.
		
----------------------

	v1 : Interrupt driven Background timer.
	v2 : polled USART routines.
	v3 : Interrupt driven USART routines.
	v4 : Now handles processing Command Strings.
	v6 : PWM working.
	v7 : Move to ATMega8 chip.
	v8 : Added Timer0
	v9 : Hardware H-Bridge connected.  Works.
   v31 : Re-configure for CD Changer Robot.
   
	v1.3	Add gantry X-Y motor controls & sensors
	v1.5	Tower Up/Down motor controls & sensors
	v1.6	Continous Cycle: Left/Right - Up/Down
			Gantry creeps back to target location on over-shoot.
	v2.0	Add 3rd Opto interrupter for center position sensor on gantry.
			Gantry & Tower now cycle thru 3 stations, lowering & raising at each station.
			Jaw assembly & Jaw relay added.  Works.  Picks up CDs.
			Add delay to simulate reading CD.
			Entire cycle now works (w/o cd reader)
			Pickup CD at station 1, drop at station 3, pause, pickup again, deposit at station 2, repeat.
	v2.1	Add support for Drawer Relay.
			Change Command String parser so cmds start with 'x' and end with carrage return.
			Add support for Drawer Limit Switch to detect when drawer is open.
			Robot works very well.  Will rip thru stack of CDs un-attended, but doesn't know when to stop.
	v2.2	Adding support to detect when stack of source CDs is empty.
	v2.3	Adding support to re-center the CD in the reader before closing the drawer.
	v2.4	Check for CD in Jaws through-out the lift
			Add more code to help the Jaw grab process
			Add 2nd push button on Dev board to add "Return to Home" function.
	v2.5	Change 2nd button to setup for iTunes to eject CD.
			Add code to force gantry to go past right edge, and creep back.
	
----------------------

	Port Pin Assignments:
	
	USART	PortD 0:1	Rx/Tx Rs-232
	Btn		PortD 2		Wired to Button on Olimex Board
	UL		PortD 3		Upper Limit Switch - Opto IR reflector sensor.
	Motor2	PortD 4:5	Motor2 Fwd/Rev Pins - outputs
	RELAY	PortD 6:7	Relay 1&2 Pins - outputs
	
	LED		PortB 0		Re-wired to LED on Olimex Board (v31)
	PWM		PortB 1:2	Motor 1&2 PWM Outputs
	Motor1	PortB 3:4	Motor1 Fwd/Rev Pins - outputs
	Btn2	PortB 5		Push-Button to Re-set Robot to Starting position 
	
	LIMIT	PortC 0:5	Input Pins for Opto-Interrupter / Limit-Switches
			PortC 0		Lower Limit Switch - Tension Micro-Switch
			PortC 1		Left Gantry Opto-Interrupter
			PortC 2		Right Gantry Opto-Interrupter
			PortC 3		Center Gantry Opto-Interrupter
			PortC 4		CD Drawer Opto-Interrupter
			PortC 5		CD Detector Micro-Switch on bottom of tower
			
************************************************/

#define F_CPU 8000000UL			// F_CPU is needed by delay.h etc

#include <avr/io.h>
#include <avr/signal.h>
#include <avr/interrupt.h>

#include "UART.h"
#include "putstr.h"
#include "motor.h"
#include "timer.h"

#define ON	1
#define OFF 0
#define RELAY_Port	PORTD

#define GANTRY_MOTOR	1
#define TOWER_MOTOR		2

#define JAW_RELAY		1
#define DRAWER_RELAY	2

#define STOPPED				0
#define MOVE_TO_FAR_RIGHT	1
#define MOVE_TO_FAR_LEFT	2
#define MOVE_TO_CENTER		3
#define CREEP_BACK			4

#define LOWER_LIMIT 0
#define LEFT_LIMIT	1
#define RIGHT_LIMIT 2
#define CENTER_LIMIT 3
#define DRAWER_LIMIT 4
#define CD_MICRO_SWITCH  5 

#define UNKNOWN		0
#define LEFT_EDGE	1
#define RIGHT_EDGE	2
#define CENTER_POS	3

#define RELAPSE_TOWER 4
#define DROP_TOWER	5
#define RAISE_TOWER 6
#define TOP_LIMIT	7
#define BOTTOM_HIT	8

#define PICKUP_CD	10
#define DROP_CD		11
#define RECENTER_CD	12

#define ROBO_CYCLE			20
#define STOP_CENTER_POS		21
#define SETUP_WAIT_EJECT	22

#define TRANSIT_SPEED	80
#define CREEP_SPEED		20
#define UP_SPEED		255
#define DOWN_SPEED		200

// --------------------------------------------------------------------------------------------------------------------------------------------------------------------------
// Prototypes

void Init_Chip(void);

unsigned char build_up_command_string(unsigned char c);
void process_command_string(void);
short command_data(unsigned char firstChr);

void turn_LED(unsigned char On_Off);
void turn_Relay(unsigned char Relay_Num, unsigned char On_Off);

void check_buttons(void);
void check_heart_beat(unsigned char *state);
unsigned char check_limit_switch(short portNum);

//void check_gantry_location(void);
void check_gantry_limits(void);
void check_lift_tower(void);
unsigned char tower_at_upper_limit(void);
unsigned char tower_at_lower_limit(void);

void Raise_Tower(void);
void Drop_Tower(void);
void Stop_Gantry(void);
void do_top_of_tower_tasks(void);

void setup_delay(unsigned char nSeconds);
void trigger_drawer_relay(void);
void wait_loop(unsigned char nSeconds);

void Timer2_Init(void);
void Timer2_OFF(void);

unsigned char we_have_a_CD(void);
unsigned char gantry_ready_to_move(void);

// --------------------------------------------------------------------------------------------------------------------------------------------------------------------------
// Static variable definitions

#define CMD_STR_SIZE 12
static unsigned char Command_String[CMD_STR_SIZE];
static unsigned char CmdStrLen = 0;
static unsigned char doing_Speed_control = 0;
static unsigned char waiting_for_iTunes = 0;
static unsigned char UL_Count = 0;
static unsigned char Max_Count = 0;
static unsigned char Stack_Empty_Count = 0;
static unsigned char we_have_seen_a_CD = 0;
static unsigned char we_should_have_cd = 0;
static unsigned char CD_Retry_Count = 0;
static unsigned char Jaw_Task = 0;
static unsigned char Robo_Task = 0;

static unsigned char Gantry_Task = 0;
static unsigned char Gantry_Location = 0;

static unsigned char Tower_Task = 0;
static unsigned char Tower_Location = 0;

static unsigned char nSecs = 0;
static unsigned char Delay_Secs = 0;
extern volatile unsigned short Timer0_ticks;
extern volatile unsigned char Timer0_10hz_Flag;

// --------------------------------------------------------------------------------------------------------------------------------------------------------------------------
// Init hardware
// This stuff is very chip dependant
// It's a pain keeping track of the port bits here -
// It's nicer when each module configs their own port setups.
// Setup for ATMega8 chip

#define XTAL 8000000UL		// 8 MHz crystal
#define BAUD 9600			// debug UART speed

void Init_Chip(void)
{
						/* Initialize port dir bits -- a '1' bit indicates an output pin*/
  DDRB = 0xDF;			/* PortB:0 is LED; PortB1:2 is PWM output. B3:4 are F/R lines. 5 is button input: 1101 1111 */
  DDRC = 0x00;			/* PortC - Sensor inputs */
  DDRD = 0xF2;			/* D0:1 is RS-232 Rx, D1 is Tx, 
						   D2:3 are sensor Input pins,
						   D4:5 are Fwd/Rev Direction Pins for Motor2,
						   D6:7 are Relay output pins. -- 0b 1111 0010 */

  PORTC = 0x3F;			/* TURN ON ALL PORT C Pullup Resistors */
  
  UART_Init(((XTAL/16)/BAUD)-1);	
  
//  UART_Init(51);		// 25 = 19.2k, 12 = 38.4k when system clock is 8Mhz (ATMega8) 
						// 51 = 9600 - XBee default
						
  Motor_PWM_Init();		/* Setup PWM using PortB1:2 on ATMega8 for output */
  
  Timer0_Init();		/* Init Tick Timer */
    
  sei();				/* Enable interrupts */

// ---

	putstr("\r\n--- Robo-CD v2.5 ---\r\n");

	turn_LED(OFF);		// LEDs are inverted. Have to explectly turn them OFF.
}

// --------------------------------------------------------------------------------------------------------------------------------------------------------------------------

void Timer2_Init(void)
{
	TCCR2 = ((1<<WGM21) | (1<<COM20) | (1<<CS20));		// setup for CTC mode, Toggle OC2 line,
	OCR2 = 0x67;										// prescaler = 1, TOP = 103.
}

void Timer2_OFF(void)
{
	TCCR2 = 0;
}

// --------------------------------------------------------------------------------------------------------------------------------------------------------------------------
// Loop forever doing the various tasks.
// Nothing requires wait states - everything cycles freely.

int main (void)
{
  unsigned char theData;
  unsigned char state = 0;
  
	Init_Chip();
	
  for (;;) {	// loop forever

	if (check_limit_switch(CD_MICRO_SWITCH))
		turn_LED(OFF);	// No CD
	else
		turn_LED(ON);	// CD Detected
	
	
	check_heart_beat( &state );		// Heart-beat is fore-ground -- true indication prog is alive.
	check_buttons();
	check_gantry_limits();
	check_lift_tower();
	
	if (waiting_for_iTunes) {
		if (check_limit_switch(DRAWER_LIMIT)) {		// returns 1 when drawer is open
			waiting_for_iTunes = 0;
			Drop_Tower();						// iTunes is done with CD - drawer just opened
			Jaw_Task = PICKUP_CD;
			we_have_seen_a_CD = 0;
			CD_Retry_Count = 0;
			}
		}
		
	if (UART_data_in_ring_buf()) {			// check for waiting UART data
		theData = UART_ring_buf_byte();		// pull 1 chr from ring buffer
		if (build_up_command_string(theData)) 
			process_command_string();		// execute commands
		}
	
	if (Timer0_10hz_Flag) {		// do these tasks only 10 times per second
		Timer0_10hz_Flag = 0;
		/*
		if (doing_Speed_control == ON)
			Motor_do_motor_control();
		*/
		}

  } // forever loop

	return 0;	// make compiler happy
} 

// --------------------------------------------------------------------------------------------------------------------------------------------------------------------------
// process incoming chars - commands start with '$' and end with '*'
// v2.1: we now start with 'x' and end with carrage return.
// return 1 if command string is complete - else return zero

unsigned char build_up_command_string(unsigned char c)
{
	if (c == 'x') {			// this will catch re-starts and stalls as well as valid commands.
		CmdStrLen = 0;
		Command_String[CmdStrLen++] = c;
		return 0;
		}
	
	if (CmdStrLen != 0)		// string has already started
		{
		if (CmdStrLen < CMD_STR_SIZE) 
			Command_String[CmdStrLen++] = c;
		return (c == 0x0D);
		}
		
	return 0;
}

// --------------------------------------------------------------------------------------------------------------------------------------------------------------------------
// we have a complete command string - execute the command.
// Command_String looks like "$A 25*" or "$s -35*" for now.
// (This code is left over from Orb prototype - RoboCD doesn't really use external commands at this time)

void process_command_string(void)
{
	// char c;
	short theData;
	unsigned char dataPos = 2;	// position of data within Command_String
	
	if (Command_String[dataPos] == ' ') dataPos++;	// skip space
	
	switch (Command_String[1]) {
					
		case 'r':	// setup ramp...
			theData = command_data(dataPos);						
			putstr("Set_Ramp: ");
			putS16( theData );
			putstr("\r\n");			
			Motor_Set_PWM( 1, 4, FORWARD );	// initial startup value
			Motor_Set_Ramp( 1, theData, 3, FORWARD );
			doing_Speed_control = ON;	// use RAMP for motor control
			break;
		
	// -----------------------
	
		case 'S':	// Stop -> All Stop
			Motor_Set_PWM( GANTRY_MOTOR, 0, 0 );
			Motor_Set_PWM( TOWER_MOTOR, 0, 0 );
			doing_Speed_control = OFF;
			putstr("STOP\r\n");
			break;

		case '?':	// Send back motor, speed, & steering info
			Motor_dump_data();
			break;

	// ---
	
		case 'K':	// Set PID Gain values  $Kp 8*
			putstr("Set Kx\r\n");
			dataPos = 3;
			if (Command_String[dataPos] == ' ') dataPos++;	// skip space
			theData = command_data(dataPos);
			
			switch (Command_String[2]) {
				case 'p': Motor_set_Kp(theData); break;
				case 'i': Motor_set_Ki(theData); break;
				case 'd': Motor_set_Kd(theData); break;
				}
							
			break;
		}
	
	CmdStrLen = 0;	// clear len, start building next command
}

// --------------------------------------------------------------------------------------------------------------------------------------------------------------------------
// scan the command string just after the command byte, 
// convert Ascii signed number to short word. (16-Bit)
// Commands look like  "$s -95*"

short command_data(unsigned char firstChr)
{
	short accum = 0;
	unsigned char sign = 0;
	unsigned char cPos = firstChr;

	if (Command_String[firstChr] == '-') {
		sign = 1;
		cPos++;
		}
	
	do {
		accum = (accum * 10) + (Command_String[cPos++] - '0');
	} while (Command_String[cPos] != '*');

	if (sign)
		accum = -accum;
	
	return accum;
}

#pragma mark-

// --------------------------------------------------------------------------------------------------------------------------------------------------------------------------
//void turn_LED(unsigned char LED_Num, unsigned char On_Off)

void turn_LED(unsigned char On_Off)
{
	unsigned char LED_Num = 0;
	
	if (On_Off == ON) {
		PORTB &= ~(1 << LED_Num); // clear pin turns LED on
		}
	else
		{
		PORTB |= (1 << LED_Num);	// set pin turns LED off
		}
}

// --------------------------------------------------------------------------------------------------------------------------------------------------------------------------
// Enter with Relay_Num in [1,2] maps to PinNum 6-7

void turn_Relay(unsigned char Relay_Num, unsigned char On_Off)
{
	unsigned char thePinNum = Relay_Num + 5;
	
	if (On_Off == ON) {
		RELAY_Port |= (1 << thePinNum);	// setting pin turns Relay ON
		}
	else
		{
		RELAY_Port &= ~(1 << thePinNum); // clearing pin turns Relay OFF
		}
}

// --------------------------------------------------------------------------------------------------------------------------------------------------------------------------

void Drop_Tower(void)
{
	Motor_Set_PWM( TOWER_MOTOR, DOWN_SPEED, FORWARD);
	Tower_Task = DROP_TOWER;		
}

void Raise_Tower(void)
{
	Motor_Set_PWM( TOWER_MOTOR, UP_SPEED, REVERSE);
	Tower_Task = RAISE_TOWER;
}

void Stop_Gantry(void)
{
	Motor_Set_PWM( GANTRY_MOTOR, 0, 0);		// Stop Motor
	Gantry_Task = STOPPED;
}

// --------------------------------------------------------------------------------------------------------------------------------------------------------------------------
// Blink Heart-Beat LED
// lets me know I'm alive - and resets Timer0
// --------------------------------------------------------------------------------------------------------------------------------------------------------------------------

void check_heart_beat(unsigned char *state)
{
//	short theData;
			
	if (Timer0_ticks > 511) {	//  512 tics per second - heart-beat led
		Timer0_reset();
		nSecs++;
	
		if (*state == 0) {
			//turn_LED(ON);
			//Timer2_Init();
			*state = 1;
			}
		else {
			//turn_LED(OFF);
			//Timer2_OFF();
			*state = 0;
			}
		
		if ((Delay_Secs > 0) && (nSecs == Delay_Secs))	// Delay is now over - CD drawer is closed
			{
			Delay_Secs = 0;
			waiting_for_iTunes = 1;		// when iTunes is done, drawer will open
			}
		
		}
}

// --------------------------------------------------------------------------------------------------------------------------------------------------------------------------

void wait_loop(unsigned char nSeconds)
{
	unsigned char t;
	for (t=0; t<nSeconds; t++)
		{
		Timer0_reset();
		while (Timer0_ticks < 512) ;
		}
	Timer0_reset();
}

// --------------------------------------------------------------------------------------------------------------------------------------------------------------------------

void setup_delay(unsigned char nSeconds)
{
	nSecs = 0;
	Delay_Secs = nSeconds;
	Timer0_reset();
}

void trigger_drawer_relay(void)
{
	Timer0_reset();
	turn_Relay( DRAWER_RELAY, ON );
	while (Timer0_ticks < 150) ;		// loop for about 1/3 second
	turn_Relay( DRAWER_RELAY, OFF );
}

// --------------------------------------------------------------------------------------------------------------------------------------------------------------------------
// Port C is setup with internal pull-up resistors on Input pins.
// Port C0:C4 = Limit Switch #0..4
// When the Opto Interrupter is clear, routine returns clear (0)
// When switch is interrupted (something in the slot blocking the light) return 1
// --------------------------------------------------------------------------------------------------------------------------------------------------------------------------

unsigned char check_limit_switch(short portNum)
{
	if (bit_is_clear(PINC, portNum))
		return 0;
	else 
		return 1;
}

// --------------------------------------------------------------------------------------------------------------------------------------------------------------------------
// When the Upper Limit sensor sees white, the Port bit is set.  (normal running position)
// When it sees black, the bit is clear. (Upper limit hit. Tower at top of travel.)
//
// Upper Limit switch is noisy.
// If we see a steady stream of 0s, then it's really at the top.
// If we ever see a 1, we're not at the top.
// Port D does not have pull-ups enabled.
// --------------------------------------------------------------------------------------------------------------------------------------------------------------------------

unsigned char tower_at_upper_limit(void)
{
	if (bit_is_clear(PIND, 3))	// does the sensor see the black mark?
		{
		if (UL_Count < 100) UL_Count++;
		if (UL_Count > Max_Count)
			{
			Max_Count = UL_Count;	// Output Debug data
			putstr("MaxCount: ");
			putU8(Max_Count);
			putstr("\r\n");
			}
		}
	else
		UL_Count = 0;
	
	return (UL_Count > 10);	// return TRUE if count > 10 - seems to work.
}

// --------------------------------------------------------------------------------------------------------------------------------------------------------------------------
// Check for slack line holding the tower

unsigned char tower_at_lower_limit(void)
	{
		if (check_limit_switch(LOWER_LIMIT))	// see if we've hit bottom
		return 0;	// line is tight - not at bottom
	else
		return 1;	// line is slack - we've hit bottom
	}

// --------------------------------------------------------------------------------------------------------------------------------------------------------------------------
// Check micro switch at bottom of tower

unsigned char we_have_a_CD(void)
	{
		if (check_limit_switch(CD_MICRO_SWITCH))
		return 0;	// No CD
	else
		return 1;	// CD Detected
	}

// --------------------------------------------------------------------------------------------------------------------------------------------------------------------------

unsigned char gantry_ready_to_move(void)
	{
	if (we_should_have_cd)
		return we_have_a_CD();
	else
		return 1;
	}

#pragma mark-

// --------------------------------------------------------------------------------------------------------------------------------------------------------------------------
//
// Tower just hit upper limit after raising the tower.
// Decide what to do: move gantry or drop tower or wait for iTunes.
//

void do_top_of_tower_tasks(void)
	{
	// check to see if we are going home.. doing Robo_Reset. 
	
	if (Robo_Task == STOP_CENTER_POS)
		{
		if (Gantry_Location == LEFT_EDGE) {
			Motor_Set_PWM( GANTRY_MOTOR, TRANSIT_SPEED, MOVE_RIGHT);
			Gantry_Task = MOVE_TO_CENTER;
			}
			
		if (Gantry_Location == RIGHT_EDGE) {
			Motor_Set_PWM( GANTRY_MOTOR, TRANSIT_SPEED, MOVE_LEFT);
			Gantry_Task = MOVE_TO_CENTER;
			}
		}

	else if (Robo_Task == SETUP_WAIT_EJECT)		// sit and wait for CD to eject
		{
		if (Gantry_Location == RIGHT_EDGE) {
			Robo_Task = ROBO_CYCLE;
			waiting_for_iTunes = 1;		// when iTunes is done, drawer will open
			}

		else {
			Motor_Set_PWM( GANTRY_MOTOR, TRANSIT_SPEED, MOVE_RIGHT);
			Gantry_Task = MOVE_TO_FAR_RIGHT;
			}
		}
	
	// ----------------------------------------
	// Doing normal cycle thru tasks ... check to see if we have a CD in our claws 

	else {		// not doing any special tasks ... just normal cycles
			
			if (we_should_have_cd)
				{
				if (!we_have_a_CD())
					{
					turn_Relay(JAW_RELAY,OFF);	// Open Jaws - 
					we_have_seen_a_CD = 0;
					if (Stack_Empty_Count < 5)
						Drop_Tower();	// go back down and try to get the CD again - else Stop.
					}
				}
			
			if (gantry_ready_to_move())		// we either have a CD, or don't need a CD.
				{
				if (Gantry_Location == LEFT_EDGE) {
					Motor_Set_PWM( GANTRY_MOTOR, TRANSIT_SPEED, MOVE_RIGHT);
					Gantry_Task = MOVE_TO_FAR_RIGHT;
					}
					
				if (Gantry_Location == RIGHT_EDGE) {	// either we just put a CD in the reader, 
					if (Jaw_Task == RECENTER_CD) {		// or removed one from the reader.
						Drop_Tower();					// or are trying to re-center it in the tray.
						Jaw_Task = DROP_CD;
						}
					else if (Jaw_Task == DROP_CD) {		// or removed one from the reader.
						trigger_drawer_relay();			// close the CD drawer - let iTunes read the CD.
						setup_delay(2);					// give the drawer some time to close
						}
					else {
						Motor_Set_PWM( GANTRY_MOTOR, TRANSIT_SPEED, MOVE_LEFT);
						Gantry_Task = MOVE_TO_CENTER;
						}
					}
					
				if (Gantry_Location == CENTER_POS) {
					Motor_Set_PWM( GANTRY_MOTOR, TRANSIT_SPEED, MOVE_LEFT);
					Gantry_Task = MOVE_TO_FAR_LEFT;
					}
				
				} // gantry_ready_to_move
			} // not doing robo_reset
			
	} // do_top_of_tower_tasks


// --------------------------------------------------------------------------------------------------------------------------------------------------------------------------
// Check all the Tower Tasks and transitions.
// Every time the tower returns to it's top position, a new gantry task begins.
//
// When the Lower Limit (line tension) sensor feels no tension, it returns zero. (Bottom hit)
// When the line is in full tension, it returns 1.
// --------------------------------------------------------------------------------------------------------------------------------------------------------------------------

void check_lift_tower(void)
{
	
	if  (Tower_Task == RELAPSE_TOWER)		// I'm not sure this ever gets called...
		{
		if (tower_at_upper_limit())		// see if we've hit the top limit switch yet.
			{
			Motor_Set_PWM( TOWER_MOTOR, 0, 0 );		// Stop Motor
			Tower_Task = STOPPED;
			Tower_Location = TOP_LIMIT;
			putstr("Relapse_Tower \r\n");
			//Drop_Tower();				// this sets tower task to drop_tower
			}
		}
	
	if (Tower_Task == DROP_TOWER)
		{
		if (tower_at_lower_limit())			// see if we've hit bottom
			{
			Motor_Set_PWM( TOWER_MOTOR, 0, 0 );		// Stop Motor
			Tower_Location = BOTTOM_HIT;
			if (Jaw_Task == PICKUP_CD)
				{
				turn_Relay(JAW_RELAY,ON);	// maybe check for CD before grabbing ? 
				we_should_have_cd = 1;		// triggering jaws may help re-center CD

				if (check_limit_switch(CD_MICRO_SWITCH))	// Check to see if there are any CDs in the stack
					Stack_Empty_Count++;	// No CD
				else
					{
					Stack_Empty_Count = 0;	// CD Detected
					we_have_seen_a_CD = 1;
					CD_Retry_Count = 0;
					}
				}
			else	// Jaw_Task is either DROP_CD or RECENTER_CD
				{
				turn_Relay(JAW_RELAY,OFF);
				we_should_have_cd = 0;
				we_have_seen_a_CD = 0;
				CD_Retry_Count = 0;
				Jaw_Task = DROP_CD;	// if Task was RECENTER, reset to DROP_CD so drawer is triggered when tower hits the top.
				}
			Raise_Tower();							// Go back up
			}		
		}
		
	if (Tower_Task == RAISE_TOWER)
		{
		if (tower_at_upper_limit())	// see if we've hit the top limit switch yet.
			{
			Motor_Set_PWM( TOWER_MOTOR, 0, 0);		// Stop Motor
			Tower_Task = STOPPED;
			Tower_Location = TOP_LIMIT;			
			do_top_of_tower_tasks();						
			}
	
		else	// we're raising the tower -- see if we've picked up and still have a CD
			{
				if (we_should_have_cd && we_have_seen_a_CD)
				if (check_limit_switch(LOWER_LIMIT))	// see if we're off the bottom 
					{
					if (check_limit_switch(CD_MICRO_SWITCH))	// Check to see if there is a CD in the jaws
						{	// No CD
						Motor_Set_PWM( TOWER_MOTOR, 0, 0);		// Stop Motor
						Tower_Task = STOPPED;
						turn_Relay(JAW_RELAY,OFF);	// Open Jaws - 
						if (CD_Retry_Count++ < 5)
							Drop_Tower();	// go back down and try to get the CD again
						else {
							turn_Relay(JAW_RELAY,OFF);
							//we_should_have_cd = 0;
							CD_Retry_Count = 0;
							we_have_seen_a_CD = 0;
							Jaw_Task = RECENTER_CD;
							Raise_Tower();				// Go back up
							// Tower_Task = RELAPSE_TOWER;
							}
						}
					}
			}
		
		} // end of if (Tower_Task == RAISE_TOWER)

}	// end of check_lift_tower()

// --------------------------------------------------------------------------------------------------------------------------------------------------------------------------
// -- Check Gantry tasks & limits --
//
// Sequence Starts at Left Edge. (Position 1)
// - Tower drops, picks up CD from stack. Raises to transit position.
// - Gantry Moves Right to Far Right Edge. (Position 3 - above CD reader)
// - Tower drops, release CD. Raises. 
// - Drop again & re-center in reader tray. raise tower.
// - ( now the cd reader should read the cd - then we pick it up again ) --

// - Move gantry to center position.
// - Drop tower. Release CD. Raise tower.
// - Move gantry to far left.  (Position 1)
// --------------------------------------------------------------------------------------------------------------------------------------------------------------------------

void check_gantry_limits(void)
{
	if (Gantry_Task == MOVE_TO_FAR_RIGHT)
		{
		if (check_limit_switch(RIGHT_LIMIT))	// Right Edge = CD reader location.
			{
			Motor_Set_PWM( GANTRY_MOTOR, CREEP_SPEED, MOVE_RIGHT);			
			do ;
			while (check_limit_switch(RIGHT_LIMIT));	// keep moving until we go past the mark
			
			Stop_Gantry();
			Drop_Tower();		
			Gantry_Location = RIGHT_EDGE;
			Jaw_Task = RECENTER_CD;				// drop the CD, then drop again to re-center in the reader tray.
			}		
		}
					
	if (Gantry_Task == MOVE_TO_CENTER)			// Center Pos = Deposit ripped CDs.
		{
		if (check_limit_switch(CENTER_LIMIT))
			{
			Stop_Gantry();
			if (Robo_Task == ROBO_CYCLE)
				Drop_Tower();
			Gantry_Location = CENTER_POS;
			Jaw_Task = DROP_CD;
			}
		}

	if (Gantry_Task == MOVE_TO_FAR_LEFT)		// We're back at the home position. Pickup another CD.
		{
		if (check_limit_switch(LEFT_LIMIT))
			{
			Motor_Set_PWM( GANTRY_MOTOR, CREEP_SPEED, MOVE_LEFT);			
			do ;
			while (check_limit_switch(LEFT_LIMIT));	// keep moving until we go past the mark

			Stop_Gantry();
			Drop_Tower();
			Gantry_Location = LEFT_EDGE;
			Jaw_Task = PICKUP_CD;
			//turn_Relay(JAW_RELAY,OFF);
			}
		}

	// ------ While dropping the tower, check the Gantry for any over-shoot.
	
	if ( (Gantry_Task == STOPPED) && (Tower_Task == DROP_TOWER) )
		{
		if (Gantry_Location == LEFT_EDGE)
			{
			if (!check_limit_switch(LEFT_LIMIT))	// check for over-shoot
				{
				Motor_Set_PWM( GANTRY_MOTOR, CREEP_SPEED, MOVE_RIGHT);
				Gantry_Task = CREEP_BACK;
				}
			}
			
		if (Gantry_Location == RIGHT_EDGE)
			{
			if (!check_limit_switch(RIGHT_LIMIT))	// check for over-shoot
				{
				Motor_Set_PWM( GANTRY_MOTOR, CREEP_SPEED, MOVE_LEFT);
				Gantry_Task = CREEP_BACK;
				}
			}			

		if (Gantry_Location == CENTER_POS)			// we got here traveling left.
			{
			if (!check_limit_switch(CENTER_LIMIT))	// check for over-shoot
				{
				Motor_Set_PWM( GANTRY_MOTOR, CREEP_SPEED, MOVE_RIGHT);
				Gantry_Task = CREEP_BACK;
				}
			}			
		}

	// ------ Creep back to the target position after over-shooting the mark.
	
	if (Gantry_Task == CREEP_BACK)
		{
		if (check_limit_switch(LEFT_LIMIT) || 
			check_limit_switch(RIGHT_LIMIT)|| 
			check_limit_switch(CENTER_LIMIT))
				Stop_Gantry();
		}
		
	// ------
}

// --------------------------------------------------------------------------------------------------------------------------------------------------------------------------
// poll the push buttons to trigger Start/Stop/Reset functions.
// --------------------------------------------------------------------------------------------------------------------------------------------------------------------------

void check_buttons(void)
{
	if (bit_is_clear(PIND, 2)) {		// Check Button on Olimex Board
		//putstr("Button...\r\n");
		turn_LED(ON);
		turn_Relay(JAW_RELAY,OFF);
		Robo_Task = ROBO_CYCLE;
		Stack_Empty_Count = 0;
		CD_Retry_Count = 0;
		
		if ((Tower_Task == STOPPED) && (Gantry_Task == STOPPED))
			{
			if (check_limit_switch(LEFT_LIMIT))		// See where we are.
				{
				Gantry_Location = LEFT_EDGE;
				Drop_Tower();
				Jaw_Task = PICKUP_CD;
				}
			else
				{
				Gantry_Location = CENTER_POS;	// Force gantry to home position.
				if (tower_at_upper_limit())
					{
					Motor_Set_PWM( GANTRY_MOTOR, TRANSIT_SPEED, MOVE_LEFT);
					Gantry_Task = MOVE_TO_FAR_LEFT;
					}
				else
					Raise_Tower();
				}
			}
				
		loop_until_bit_is_set(PIND, 2);
		turn_LED(OFF);
		}

	if (bit_is_clear(PINB, 5)) {		// Check Button 2 on Olimex Board
		//putstr("Button 2...\r\n");
		turn_LED(ON);
		turn_Relay(JAW_RELAY,OFF);
		we_should_have_cd = 0;
		Robo_Task = SETUP_WAIT_EJECT;	// Travel to Center Position and Stop.
		
		if (tower_at_upper_limit())
			{
			if (Gantry_Location != RIGHT_EDGE)		// See where we are.
				{
				Motor_Set_PWM( GANTRY_MOTOR, TRANSIT_SPEED, MOVE_RIGHT);
				Gantry_Task = MOVE_TO_FAR_RIGHT;
				}
			else 
				{
				Robo_Task = ROBO_CYCLE;
				waiting_for_iTunes = 1;		// when iTunes is done, drawer will open
				}
			}
		else
			Raise_Tower();		
		
		loop_until_bit_is_set(PINB, 5);
		turn_LED(OFF);
		}
}

// --------------------------------------------------------------------------------------------------------------------------------------------------------------------------
// End of File
