// motor.c - Motor Electronic Speed Controller Functions
//
// Output 2 channels of PWM on PortB Pins 1&2
// Ouput 2 sets of control lines (Fwd/Rev)
// 2 motors are setup as PID controlled continous rotation motors

#include <avr/io.h>
#include "UART.h"
#include "putstr.h"
#include "motor.h"

// These are chip dependant port pins
// For ATMega8 OC1A is PortB1, OC1B is PortB2
// Forward / Reverse Direction Bits are Port B3:4 & D4:5

#define MC1_Fwd_Pin 0x08			// b 0000 1000		PortB:3
#define MC1_Rev_Pin 0x10			// b 0001 0000		PortB:4
#define MC1_Port	PORTB

#define MC2_Fwd_Pin 0x10			// b 0001 0000		PortD:4
#define MC2_Rev_Pin 0x20			// b 0010 0000		PortD:5
#define MC2_Port	PORTD

#define MC1_PWM_MASK 0x02			// b 0000 0010		PortB1
#define MC2_PWM_MASK 0x04			// b 0000 0100		PortB2

#define	mcNONE	0
#define	mcRAMP	1
#define mcPID	2

// ------------------------------------------------------------------------------------------------------------------------------------------------------
// Motor Control Blocks for keeping track of Motor PID function variables

typedef struct {
	char	mDriveNum;
	char	Kp;
	char	Ki;
	char	Kd;
	char	dir;
	char	PID_state;
	char	mcFunc;
	unsigned short	target_value;	
	unsigned short	last_set_point;	
	unsigned short	prev_value;
	unsigned short	crnt_value;
	short	ramp_inc;
	short	power_setting;
	short	PWM_Set;
	short	I_State;
	short	D_State;
	short	initialError;
} motor_control_block;

/* Static Vars */
static motor_control_block motor1, motor2;

/* Prototype */
void Motor_clear_mcb( motor_control_block *m );
short calc_motor_drive(motor_control_block *motor);

// ------------------------------------------------------------------------------------------------------------------------------------------------------
// Special care must be taken writing to these 16 bit PWM registers.
// High byte must be written first.

void write_OCR1A( unsigned char value )
{
	OCR1AH = 0;
	OCR1AL = value;
}

void write_OCR1B( unsigned char value )
{
	OCR1BH = 0;
	OCR1BL = value;
}

// ------------------------------------------------------------------------------------------------------------------------------------------------------
// Setup torque control - turn on PID

void Motor_Set_Torque(unsigned char motorNum, unsigned char t, signed char direction)
{
	if (motorNum == 1) {
		motor1.target_value = t;
		motor1.dir = direction; 
		motor1.mcFunc = mcPID; }
	else {
		motor2.target_value = t;
		motor2.dir = direction;
		motor2.mcFunc = mcPID;
		}
}

// ------------------------------------------------------------------------------------------------------------------------------------------------------

void Motor_Set_Ramp(unsigned char motorNum, unsigned short targetValue, short rampInc, signed char direction)
{
	motor_control_block *m;
	
	if (motorNum == 1) m = &motor1;
	else m = &motor2;
		
	m->target_value = targetValue;
	m->ramp_inc = rampInc;
	m->dir = direction;
	m->mcFunc = mcRAMP;
}

// ------------------------------------------------------------------------------------------------------------------------------------------------------
// Set power -100..0..100 as percent of PWM (from cmd line)

void Motor_Set_Power(unsigned char motorNum, unsigned char power, signed char direction)
{
	short tmpData;
	tmpData = (power * 255) / 100;
	if (tmpData > 255) tmpData = 255;

	if (motorNum == 1) {
		Set_Motor1_PWM( tmpData, direction);
		motor1.power_setting = power; }
	else {
		Set_Motor2_PWM( tmpData, direction);
		motor2.power_setting = power;
		}
}

void Motor_Set_PWM(unsigned char motorNum, unsigned char pwm, signed char direction)
{
	if (motorNum == 1) 
		Set_Motor1_PWM( pwm, direction);
	else
		Set_Motor2_PWM( pwm, direction);
}

// ------------------------------------------------------------------------------------------------------------------------------------------------------
// Sets the Duty Cycle and direction of motor 1 - Main Drive Motor
// This routine is tied to specific Speed Control Hardware
// input value is 0..255

void Set_Motor1_PWM(unsigned char pwm, signed char direction)
{	
	if (pwm == 0)		// STOP - Turn off PWM
		{
		write_OCR1A( 0 );
		MC1_Port &= ~(MC1_Fwd_Pin | MC1_Rev_Pin);		// clear both pins
		}
	else
		{		
		if (direction == FORWARD)
			{
			write_OCR1A( pwm );			// Set PWM as 16 bit value
			MC1_Port |= MC1_Fwd_Pin;	// Set pin
			MC1_Port &= ~MC1_Rev_Pin;	// Clear pin
			}
		else // direction == REVERSE
			{
			write_OCR1A( pwm );			// Set PWM as 16 bit value
			MC1_Port &= ~MC1_Fwd_Pin;	// Clear pin
			MC1_Port |= MC1_Rev_Pin;	// Set pin
			}
		}
		
	motor1.dir = direction;
	motor1.PWM_Set = pwm;
}

// ------------------------------------------------------------------------------------------------------------------------------------------------------
// Sets the Duty Cycle and direction of motor 2 - 
// pwm = 0..255

void Set_Motor2_PWM(unsigned char pwm, signed char direction)
{
	if (pwm == 0)		// STOP - Turn off PWM
		{
		write_OCR1B( 0 );
		MC2_Port &= ~(MC2_Fwd_Pin | MC2_Rev_Pin);
		}
	else
		{		
		if(direction == FORWARD)
			{
			write_OCR1B( pwm );			// Set PWM as 16 bit value
			MC2_Port |= MC2_Fwd_Pin;	// Set direction pins
			MC2_Port &= ~MC2_Rev_Pin;
			}
		else // direction == REVERSE
			{
			write_OCR1B( pwm );			// Set PWM as 16 bit value
			MC2_Port &= ~MC2_Fwd_Pin;	// Clear pin
			MC2_Port |= MC2_Rev_Pin;	// Set pin
			}
		}
	motor2.dir = direction;
	motor2.PWM_Set = pwm;
}

// ------------------------------------------------------------------------------------------------------------------------------------------------------
// Init Pulse Width Modulation hardware for Speed Controllers.
// Use 16 bit Counter/Timer1 on ATMega8
// Output 2 channels of PWM on PortB Pins 1&2
// Ouput 2 sets of control lines (Fwd/Rev) on ports B & D
// Main Drive motor is setup as PID controlled continous rotation (drive)
// Steering Motor is setup as a Servo with a position sensing feedback pot.

void Motor_PWM_Init(void)
{
	Motor_clear_mcb( &motor1 );
	Motor_clear_mcb( &motor2 );
	
	motor1.mDriveNum = 1;
	motor2.mDriveNum = 2;
	
    // Set Port Direction bits (1=Output) and enable PWM pins and timers
	
    DDRB |= MC2_PWM_MASK | MC1_PWM_MASK;
    DDRB |= MC1_Fwd_Pin | MC1_Rev_Pin;    
    DDRD |= MC2_Fwd_Pin | MC2_Rev_Pin;    

	TCCR1A |= (1<<COM1A0) | (1<<COM1A1);	// Set OC1A on compare match
    TCCR1A |= (1<<COM1B0) | (1<<COM1B1);	// Set OC1B on compare match
	
    TCCR1A |= (1<<WGM10);		// Fast PWM Mode 5 ==> 8 Bit
	TCCR1B |= (1<<WGM12);
	
    TCCR1B |= (1<<CS11);		// 8 prescale = 1.2khz PWM @3.6MHz
 
    // Make sure motors are stopped
    Set_Motor1_PWM(0, FORWARD);
    Set_Motor2_PWM(0, FORWARD);
}

// ------------------------------------------------------------------------------------------------------------------------------------------------------
// Initialize the motor control blocks

void Motor_clear_mcb( motor_control_block *m )
{
	m->dir = FORWARD;
	m->target_value = 0;
	m->prev_value = 0;
	m->ramp_inc = 0;
	
	m->PWM_Set = 0;
	m->I_State = 0;
	m->D_State = 0;

// iTerm PID
	m->Kp = 8;
	m->Ki = 10;
	m->Kd = 4;
	
/* normal PID	
	m->Kp = 4;
	m->Ki = 1;
	m->Kd = 4;
*/
	m->PID_state = 0;
	m->mcFunc = mcNONE;
	m->mDriveNum = 0;
}

// ------------------------------------------------------------------------------------------------------------------------------------------------------

char Motor_Read_Drive_Direction(void)
{
	return motor1.dir;
}

// ------------------------------------------------------------------------------------------------------------------------------------------------------

void Motor_Set_Drive_Direction(char theDir)
{
	motor1.dir = theDir;
}

// ------------------------------------------------------------------------------------------------------------------------------------------------------
// Setup the Gain factors via the Cmd line for testing & tuning.

void Motor_set_Kp(char c)
{
	motor1.Kp = c;
}

void Motor_set_Ki(char c)
{
	motor1.Ki = c;
}

void Motor_set_Kd(char c)
{
	motor1.Kd = c;
}

// ------------------------------------------------------------------------------------------------------------------------------------------------------
// Dump motor data to serial port for feedback / debug / testing / tuning

void Motor_dump_data(void)
{
	putstr("Motor1: ");
	putS16(motor1.target_value);
	putS16(motor1.crnt_value);
	putS16(motor1.PWM_Set);
	putstr("  PID:");
	putS16(motor1.Kp);
	putS16(motor1.Ki);
	putS16(motor1.Kd);
	putstr("  ISav:");
	putS16(motor1.I_State);	
	putstr("\r\n");
}

// ------------------------------------------------------------------------------------------------------------------------------------------------------

short calc_motor_drive(motor_control_block *motor)
{
//	short Error_Term;
	short motor_drive;
	motor_drive = motor->PWM_Set;
	
	if (motor->mcFunc == mcRAMP)
		{
		motor->crnt_value = motor->PWM_Set;
		if (motor->crnt_value != motor->target_value)
			{
			motor_drive = motor->crnt_value + motor->ramp_inc;
			if (motor_drive > motor->target_value)
				motor_drive = motor->target_value;
			}
		}
	
	// -----
	
	else if (motor->mcFunc == mcPID)
	{
//	motor1.crnt_value = Encoder_read_speed(MOTOR1_SHAFT_ENCODER);
	motor->crnt_value = motor->PWM_Set;
	motor_drive = motor->target_value;		
	}
	
	if (motor_drive > 255) motor_drive = 255;
	if (motor_drive < 0) motor_drive = 0;
	
	return motor_drive;
	
}  // calc_motor_drive


// ------------------------------------------------------------------------------------------------------------------------------------------------------

void Motor_do_motor_control(void)
{
	short motor_drive;
	
	motor_drive = calc_motor_drive( &motor1 );
	if (motor_drive > 160) motor_drive = 160;		// limit while testing
	if (motor_drive != motor1.PWM_Set)
		Set_Motor1_PWM( motor_drive, motor1.dir );	// PWM input is 0..255  50% = 128

	motor_drive = calc_motor_drive( &motor2 );
	if (motor_drive > 160) motor_drive = 160;		// limit while testing
	if (motor_drive != motor2.PWM_Set)
		Set_Motor2_PWM( motor_drive, motor2.dir );	// PWM input is 0..255  50% = 128
}

// ------------------------------------------------------------------------------------------------------------------------------------------------------
// End of File
